/***************************************************************************
 *   Copyright (C) 2007 by Harm van Eersel                                 *
 *   Copyright (C) 2009 by Tim Vandermeersch                               *
 *   Copyright (C) 2009 by Nicola Zonta                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/

#ifndef MSK_MOLSCENE_H
#define MSK_MOLSCENE_H

#include <QGraphicsScene>
#include <QUndoCommand>

#include "bond.h"
#include "abstractxmlobject.h"
#include "stringify.h"

class QString;
class QImage;
class QListWidgetItem;
class QTableWidgetItem;
class QUndoStack;

namespace Molsketch {

  class Molecule;
  class Atom;
  class Bond;
  class TextInputItem;
  class genericAction;
  class SceneSettings;

  class MolScene : public QGraphicsScene, public abstractXmlObject
  {
    Q_OBJECT

    public:
    static const QString mouseWheelForCyclingTools;
      enum EditMode {
        MoveMode,     //!< Atoms and molecules are movable.
        DrawMode,     //!< Mode to add atoms and bonds.
        RotateMode,   //!< Mode to rotate molecules.
        ChargeMode,   //!< Increase/decrease charges
        HydrogenMode, //!< increase/decrease implicit hydrogen count
        LassoMode,    //!< lasso selection tool
        TextMode,     //!< add / edit text
        MinimiseMode, //!< gemoetry optimise
        ConnectMode,  //!< mode to drag/drop items
        ReactionMode, //! mode to insert reaction arrows
        MechanismMode, //!< mode for inserting mechanism arrows
      };

      enum RenderMode
      {
        RenderLabels,
        RenderColoredSquares,
        RenderColoredCircles,
        RenderColoredWireframe,
      };

      explicit MolScene(QObject* parent = 0);
      /** Creates a new MolScene with @p settings. Will assume ownership of @p settings! */
      MolScene(SceneSettings *settings, QObject* parent = 0);
      ~MolScene();

      SceneSettings* settings() const;

      QFont getAtomFont() const;

      static QString mimeType();
      int editMode() const; // TODO obsolete?

      qreal bondAngle() const;

      /**
       * @return The current RenderMode. Default is RenderLabels
       */
      RenderMode renderMode() const; // TODO do we really need this?
      /**
       * Set the RenderMode.
       */
      void setRenderMode(RenderMode mode);

      // Commands
      /** Renders the @p rect on the scene in a image. */
      QImage renderImage(const QRectF &rect);

      QByteArray toSvg();


      QUndoStack *stack() const;

      QList<Molsketch::genericAction *> sceneActions() const;

      // Advanced queries
      /** Returns the first molecule at position @p pos or NULL if none. */
      Molecule* moleculeAt(const QPointF &pos);
      /** Returns the first atom at position @p pos or NULL if none. */
      Atom* atomAt(const QPointF &pos);
      /** Returns the first bond at position @p pos or NULL if none. */
      Bond* bondAt(const QPointF &pos);

      /** Calculates the nearest magnetic point around @p curPos. */
      QPointF nearestPoint(const QPointF &curPos);

      TextInputItem *inputItem() ;

      QList<Atom*> atoms() const;

    signals:
      void copyAvailable(bool);
      void pasteAvailable(bool);
      void editModeChange(int);
      void documentChange( );

    public slots:
      /** Slot to cut the current selection to the clipboard. */
      void cut();
      /** Slot to copy the current selection to the clipboard. */
      void copy();
      /** Slot to paste the current clipboard contents. */
      void paste();
      /** Slot to clear the scene. */
      void clear();
      /** Slot to set the edit mode to @p mode. */
      void setEditMode(int mode);
      /** Slot to select all contents of the scene. */
      void selectAll();
      void addMolecule(Molecule* mol);
      void setGrid(bool on = true);
      void clipboardChanged();


#ifdef QT_DEBUG
  void debugScene();
#endif


    protected:
      void keyPressEvent(QKeyEvent* keyEvent);
      void contextMenuEvent(QGraphicsSceneContextMenuEvent *event);
      void mousePressEvent(QGraphicsSceneMouseEvent *event);
      void mouseMoveEvent(QGraphicsSceneMouseEvent *event);
      void mouseReleaseEvent(QGraphicsSceneMouseEvent *event);

      void dragEnterEvent(QGraphicsSceneDragDropEvent *event);
      void dragLeaveEvent(QGraphicsSceneDragDropEvent *event);
      void dropEvent(QGraphicsSceneDragDropEvent *event);
      void dragMoveEvent(QGraphicsSceneDragDropEvent *event);

      void wheelEvent(QGraphicsSceneWheelEvent *event);
  public:

      QString xmlName() const;
      static QString xmlClassName();
      QPointF snapToGrid(const QPointF& point, bool force = false);
      bool snappingToGrid() const;
      bool cyclingByMouseWheelEnaled() const;
      Atom *atomNear(const QPointF &pos, qreal tolerance = 10.0);
      template <class T>
      T *itemNear(const QPointF &pos, qreal tolerance);
      qreal getRadicalDiameter() const;
      qreal getLonePairLength() const;
      qreal getLonePairLineWidth() const;
      QWidget *getPropertiesWidget(); // TODO there should be an intgerface for this
  protected:
      XmlObjectInterface *produceChild(const QString &childName, const QString& type) ;
      QList<const XmlObjectInterface *> children() const ;
      void readAttributes(const QXmlStreamAttributes &attributes) ;
      QXmlStreamAttributes xmlAttributes() const;

  private:

      // Hinting methods and properties
      /** Shows a highlight rectangle around @p item. */
      void setHoverRect( QGraphicsItem* item );

      /** Stores the edit mode of the scene as an integer. */
      int m_editMode;
      RenderMode m_renderMode;

      class privateData;
      privateData *d;

      // Auxillary feedback items
      /** The highlight rectangle. */
      QGraphicsPathItem* m_hoverRect;

  private slots:
      void updateAll() ;
      void selectionSlot();
      void updateGrid(const QRectF &newSceneRect);
  };

} // namespace

#endif
